/**************************************************************************************************************
 *
 * PIC18F6620-Modul
 *
 * Datenlogger fr Raketenwaage
 *
 * Datei : Hello.c
 *
 * Autor : Finger
 *
 * Controller : Microchip 18F6620, 20 Mhz
 * Einstellungen:  OSC=HS, WDT=OFF, PWT=ON, CP=OFF, BOD=ON, LVP=OFF
 *
 * TAB-Space : 2
 * Editor : MI-Edit
 * Entwicklungsumgebung : Hi-Tech PIC-C 18 (www.htsoft.com)
 *
 *************************************************************************************************************/

#include	<pic18.h>
#include	<stdlib.h>
#include	<stdio.h>
#include  <string.h>
#include  "defines.h"

//----------------------------------------------------------
// Rckgabepuffer fr INT -> ASCII-Wandlung und Textausgabe
//----------------------------------------------------------
char outpbuf [Buflen];
char Puffer [20];

//-----------
// Mentexte
//-----------
char HauptMenueText [HauptMenueAnzahl] [LCDZeile+1] =
 {
  "Slot 1 aufz.   \0",
  "Slot 2 aufz.   \0",
  "Slot 3 aufz.   \0",
  "Slot 4 aufz.   \0",
  "Slot 1 Analyse \0",
  "Slot 2 Analyse \0",
  "Slot 3 Analyse \0",
  "Slot 4 Analyse \0",
  "Aktueller Wert \0",
  "Kalibration    \0"
 };

char HilfeText1 [LCDZeile] = "<Nein>     <Ja>";
char HilfeText2 [LCDZeile] = "<ESC>   <Start>";
char HilfeText3 [LCDZeile] = "Beendet    <OK>";
char HilfeText4 [LCDZeile] = "Max        <OK>";
char HilfeText5 [LCDZeile] = "Rennt....      ";
char HilfeText6 [LCDZeile] = "           <OK>";
char HilfeText7 [LCDZeile] = "Punkt 1 : __._N";
char HilfeText8 [LCDZeile] = "<->   <+>  <OK>";
char HilfeText9 [LCDZeile] = "Sicher ?   <OK>";

//-----------------
// Zeitsteuerungen
//-----------------
volatile unsigned long Systemticker = 0;
volatile unsigned long Ticker = 0;
volatile unsigned int Sekundenzaehler = 0;
volatile unsigned int Sekundenticker = 0;
volatile unsigned int SekundentickerAlt = 0;
unsigned long Pauser;

//--------------
// Aufzeichnung
//--------------
volatile char Slotnummer = 0;
char Messwerte [64];
char DevAdresse = 0;
unsigned int Adresse = 0;
char RAMZeiger = 0;
int Messpuffer [Aufzeichnung];                    // Speichern der Messwerte im RAM
int MessPufferZeiger = 0;
char *Zeiger;
int i;

//---------
// Display
//---------
static bit LCD_RS	@ ((unsigned)&PORTD*8+2);				// LCD: Register select
static bit LCD_RW	@ ((unsigned)&PORTD*8+3);				// LCD: Read / Write
static bit LCD_EN0	@ ((unsigned)&PORTE*8+0);			// Enable 0

char Zeile = 0;
char Spalte = 0;

//---------------------
// Geradenberechnungen
//---------------------
float Nenner;
float Zaehler;
float Steigung;
float Nullpunkt;
float Kraft;

int IntKraft;
int VorkommaIntKraft;
int NackkommaIntKraft;

//--------------------
// Kalibrationspunkte
//--------------------
int x1;
int x2;
int y1;
int y2;

char Minus;

//---------------------------------
// Konfiguration des Controllers
//---------------------------------
__CONFIG(1, OSCSEN & HS);                         // Quarz 20 Mhz
__CONFIG(2, BORDIS & WDTDIS);                     // Kein Watchdog, kein BrownOut
__CONFIG(4, LVPDIS & STVREN);                     // Kein LowVoltageProgramm, Reset bei StackOverflow

/***************************************************************************************************************
 *
 * Kleine Pause
 *
 ***************************************************************************************************************/
void DelayUs (int x)
 {
  unsigned int _dcnt;

	_dcnt = x;
	while(--_dcnt != 0) continue;
 }

/***************************************************************************************************************
 *
 * Verzgern im Millisekundebereich
 *
 ***************************************************************************************************************/
void Delayms (unsigned int cnt)
 {
   unsigned char	i;
  	do {
      		i = 1;
      		do {
      			   DelayUs (1000);
      		   } while(--i);
       } while(--cnt);
 }

/******************************************************************************
 *
 *  Internes EEProm beschreiben fr PIC 18F6720
 *
 *******************************************************************************/
void EEPWrite (unsigned int ADR, char DATA)
 {
  EEADRH = (ADR>>8) & 0x3;
  EEADR = ADR & 0xFF;     // Adresse
	EEDATA = DATA;          // Daten
	EEPGD = 0;              // EEProm beschreiben
	CFGS = 0;
	WREN = 1;               // Schreiben freigeben

	GIE = 0;                // globaler interrupt aus
	EECON2 = 0x55;          // ist vorgeschrieben mu so gemacht werden damit wird die
	EECON2 = 0xAA;          // Speicherung gestartet
	WR = 1;                 // Schreiben
	while (WR==1);          // warten bis fertig, WR wird zurck gesetzt wenn der
	                        // Schreibzyklus abgeschlossen ist
	GIE = 1;                // globaler interrupt an
	WREN = 0;               // Schreiben sperren
 }

/******************************************************************************
 *
 *  Internes EEProm lesen
 *
 *******************************************************************************/
char EEPRead (unsigned int ADR)
 {
  EEADRH = (ADR>>8) & 0x3;
  EEADR = ADR & 0xFF;     // Adresse
	EEPGD = 0;
	CFGS = 0;
	RD = 1;
	return EEDATA;
 }


/***********************************************************************************
 *
 * Macht den AD-Wandler startklar
 *
 ***********************************************************************************/
void adc_init (void)
 {
  ADCON0 = 0b00000001;        // Kanal 0, ADC ein
  ADCON1 = 0b00110000;        // Externe Referenzen, AN0 Analogeingang
  ADCON2 = 0b10000110;        // Right justify, Clock = Fosc/64
  TRISA |= 0b00000111;        // Bit 0-2 : Analog in
 }

/***********************************************************************************
 *
 * Liest einen "rohen" Wandlerwert ein
 *
 ***********************************************************************************/
int adc_get_raw (void)
 {
  int tempwert = 0;

  GIE = 0;

  ADIF = 0;                           // Interrupt-Flag rcksetzen
  ADCON0 |= 0x02;                     // Wandler starten
  while (ADCON0 & 0x02);              // Warten, bis fertig
  while (!ADIF);

  tempwert = ADRESH*256 + ADRESL;

  GIE = 1;

  return (tempwert);       // Wert zurckliefern

 }

/**************************************
 *
 * Liefert einen Kraftwert * 10 zurck
 *
 **************************************/
int adc_get_newton (void)
 {

  int wert;

  // Punkte aus dem EEP laden
  x1 = EEPRead (0) * 256;
  x1 += EEPRead (1);
  y1 = EEPRead (2) * 256;
  y1 += EEPRead (3);
  x2 = EEPRead (4) * 256;
  x2 += EEPRead (5);
  y2 = EEPRead (6) * 256;
  y2 += EEPRead (7);

  // Korrekturwerte setzen
  if (x1 > 1023) x1 = 10;
  if (y1 > 150) y1 = 100;
  if (x2 > 1023) x2 = 1000;
  if (y2 > 150) y2 = 100;

  // Berechnung der Geradenparameter
  Nenner = (y2 - y1) / 10;
  Zaehler = x2 - x1;
  Steigung = Nenner / Zaehler;
  Nullpunkt = (float)(y2);
  Nullpunkt /= 10;
  Nullpunkt -= Steigung*x2;

  // Kraft bestimmen
  wert = adc_get_raw ();
  Kraft = Steigung * wert + Nullpunkt;
  return ((int)(Kraft*10));

 }

/*****************************************************************************************************************
 *
 * String nach USART schreiben, nicht ISR-gesteuert
 *
 *****************************************************************************************************************/
void print (char UARTNo, const char *data)
{
  char cdummy;                                   // Dummyaufnahme des Empfangsregister

  if (UARTNo == 0)
   { // Ersten UART ansprechen
    if (FERR1) cdummy = RCREG1;
    if (OERR1)
    	{
      	CREN1=0;
      	CREN1=1;
    	}
    	do
    	 {
    		while (!TRMT1);
      	TXREG1 = *data;
      	data++;
    	 } while ( *data );
   } else // Von if (UARTNo == 0)

   { // Zweiten UART ansprechen
    if (FERR2) cdummy = RCREG2;
    if (OERR2)
    	{
      	CREN2=0;
      	CREN2=1;
    	}
    	do
    	 {
    		while (!TRMT2);
      	TXREG2 = *data;
      	data++;
    	 } while ( *data );

   } // Von else if (UARTNo == 0)

}

/***********************************************************************************
 *
 * Schaltet SDA und SCL als Eingang und macht die Schnittstelle startklar
 *
 ***********************************************************************************/
void i2c_init (void)
 {
  TRISC |= SDA_PORT;                                             // SDA = RC4 = Input
  TRISC |= SCL_PORT;                                             // SCL = RC3 = Input

  SSPCON1 = 0b00001000;                                           // Master-Mode, Clock : SSPADD
  SSPEN = 1;                                                      // Enable
  SSPIF = 0;                                                      // Interruptflag rcksetzen
  SSPADD = 0x6F;                                                  // I2C-Clock

 }

/*************************************************************************************
 *
 * Erzeugt die Stop-Condition
 *
 *************************************************************************************/
void i2c_close (void)
 {

  unsigned long timeout;

  PEN = 1;                                                  // Start-Condition generieren
  while (PEN);                                              // Warten, bis die Stop-Condition fertig ist

  timeout = Systemticker;
  while ((!SSPIF) && ( (timeout+10) > Systemticker ) );

  if ( (timeout+10) <= Systemticker ) print (UART1, " NOCH IMMER NIX SSPIF");
  if (BCLIF) print (UART1, " BCLIF=1 ");

  SSPIF = 0;                                                // Flag rcksetzen
       /*
  SSPEN = 0;
  NOP();NOP();NOP();NOP();NOP();NOP();NOP();NOP();
  SSPEN = 1;
     */
  // while (1);

 }

/*************************************************************************************
 *
 * Adressiert einen Slave, liefert TRUE falls dieser mit ACK antwortet, sonst FALSE
 * Richtung TRUE = Schreiben zum Slave, FALSE = Lesen vom Slave
 *
 *************************************************************************************/
char i2c_open (char Adresse, char Richtung)
 {
   unsigned long timeout;
 // print (UART1, "T");

/*
i2c_close ();
SSPEN = 0;
NOP();NOP();NOP();NOP();NOP();NOP();NOP();NOP();
SSPEN = 1;
*/
  if (BCLIF) print (UART1, " BCLIF=1 ");
       /*
  sprintf (outpbuf, "\n\rSSPSTAT : %d\n\r", SSPSTAT);
  print (UART1, outpbuf);
  sprintf (outpbuf, "SSPCON2 : %d\n\r", SSPCON2);
  print (UART1, outpbuf);
     */
  SSPEN = 0;
  // print (UART1, "Aus ");
  SSPEN = 1;
  // print (UART1, " An \n\r");

  SEN = 1;                                                  // Start-Condition generieren
         /*
  sprintf (outpbuf, "SSPSTAT : %d\n\r", SSPSTAT);
  print (UART1, outpbuf);
  sprintf (outpbuf, "SSPCON2 : %d\n\r", SSPCON2);
  print (UART1, outpbuf);
           */
  if (BCLIF) print (UART1, " BCLIF=1 ");

  timeout = Systemticker;
  while ((!SSPIF) && ( (timeout+10) > Systemticker ) );
   //print (UART1, "+");                                      // Warten, bis die Start-Condition fertig ist

  if (BCLIF) print (UART1, " BCLIF=1 ");

  if ( (timeout+10) <= Systemticker ) print (UART1, " NIX SSPIF");

  SSPIF = 0;                                                // Flag rcksetzen

  if (Richtung) SSPBUF = Adresse; else SSPBUF = Adresse+1;  // Adresse senden

  timeout = Systemticker;
  while ((!SSPIF) && ( (timeout+10) > Systemticker ) );
   // print (UART1, "#");                        // Warten auf das setzen des Interrupt-Flags

  if (BCLIF) print (UART1, " BCLIF=1 ");

  if ( (timeout+10) <= Systemticker ) print (UART1, " WIEDER NIX SSPIF");

  SSPIF = 0;                                                // Flag rcksetzen

  if (!(ACKSTAT)) return (TRUE); else                       // Prfen auf ACK vom Slave
                  return (FALSE);
 }

/*************************************************************************************
 *
 * Schreibt ein Byte via i2C, liefert TRUE bei ACK vom Slave, sonst FALSE
 *
 *************************************************************************************/
char i2c_writebyte (unsigned char wert)
 {
  SSPBUF = wert;

  while (!SSPIF);                                           // Warten auf das setzen des Interrupt-Flags
  SSPIF = 0;                                                // Flag rcksetzen

  if (!(ACKSTAT)) return (TRUE); else                       // Prfen auf ACK vom Slave
                  return (FALSE);

 }

/***********************************************************************************
 *
 * Liest ein Byte via i2C, erzeugt bei ACK ein Acknowledge
 *
 ***********************************************************************************/
char i2c_readbyte (char ACKn)
 {
   RCEN = 1;                                  // Receive einschalten
   while (RCEN);
   ACKEN = ACKn;                              // Acknowledgeerzeugung bei Bedarf einschalten
   while (!SSPIF);                            // Warten auf das setzen des Interrupt-Flags
   SSPIF = 0;                                 // Flag rcksetzen
   return (SSPBUF);                           // Wert bergeben
 }

/*********************************************************************
 *
 * Liest 1 Byte aus dem I2C-EEPROM an der bergebenen Adresse
 *
 *********************************************************************/
char i2c_EEPROM_Read_Byte (char dev, unsigned int Adresse)
 {
  char b;
  char tester;

  // Slave adressieren zum Schreiben
  i2c_open (dev, WRITE);

  // Adresse schreiben
  // High-Byte
  b = (Adresse & 0xff00) >> 8;     // Highbyte ausmaskieren
  i2c_writebyte (b);

  // Low-Byte
  b = (Adresse & 0xff);            // Lowbyte ausmaskieren
  i2c_writebyte (b);

  // Verbindung beenden
  i2c_close ();

  // Slave adressieren zum Lesen
  i2c_open (dev, READ);

  // Datum auslesen
  tester = i2c_readbyte (FALSE);

  // Verbindung beenden
  i2c_close ();

  return (tester);
 }

/*********************************************************************
 *
 * Schreibt 1 Byte in das I2C-EEPROM an der bergebenen Adresse
 * Liefert TRUE bei OK
 *
 *********************************************************************/
char i2c_EEPROM_Write_Byte (char dev, unsigned int Adresse, char Datum)
 {
  char b;
  char tester;

  // Slave adressieren zum Schreiben
  tester = i2c_open (dev, WRITE);

  if (!(tester))
   {
    // Verbindung beenden
    i2c_close ();
    return (FALSE);
   }

  // Adresse schreiben
  // High-Byte
  b = (Adresse & 0xff00) >> 8;     // Highbyte ausmaskieren
  i2c_writebyte (b);

  // Low-Byte
  b = (Adresse & 0xff);            // Lowbyte ausmaskieren
  i2c_writebyte (b);

  // Datum
  i2c_writebyte (Datum);

  // Verbindung beenden
  i2c_close ();

  Delayms (10);

  // berprfen, wann der Schreibzyklus abgelaufen ist. Erst dann gibt es wieder ein ACK
  tester = FALSE;
  while (!(tester))
   {
    tester = i2c_open (dev, WRITE);
    DelayUs (50);
    i2c_close ();
    DelayUs (50);
   }

  return (TRUE);
 }

/*********************************************************************
 *
 * Liest n Bytes aus dem I2C-EEPROM an der bergebenen Adresse
 * Liefert FALSE, falls ein Fehler aufgetreten ist
 *
 *********************************************************************/
char i2c_EEPROM_Read (char dev, unsigned int Adresse, char laenge, char* puffer)
 {
  char b;
  char puf1 [Buflen];
  char zeiger;

  memset (puf1, 255, Buflen);

  // Slave adressieren zum Schreiben
  if (!(i2c_open (dev, WRITE)))
   {
     // Verbindung beenden
     i2c_close ();
     return (FALSE);
   }

  // Adresse schreiben
  // High-Byte
  b = (Adresse & 0xff00) >> 8;                            // Highbyte ausmaskieren

  if (!(i2c_writebyte (b)))
   {
     // Verbindung beenden
     i2c_close ();
     return (FALSE);
   }

  // Low-Byte
  b = (Adresse & 0x00ff);                                 // Lowbyte ausmaskieren

  if (!(i2c_writebyte (b)))
   {
     // Verbindung beenden
     i2c_close ();
     return (FALSE);
   }

  // Verbindung beenden
  i2c_close ();

  // Slave adressieren zum Lesen
  if (!(i2c_open (dev, READ)))
   {
     // Verbindung beenden
     i2c_close ();
     return (FALSE);
   }

  // Daten auslesen
  zeiger = 0;
  while (zeiger < laenge)
   {
    if ((zeiger+1) == laenge) puf1 [zeiger] = i2c_readbyte (FALSE); else
                              puf1 [zeiger] = i2c_readbyte (TRUE);
    // Ein bischen Zeit verbummeln
    DelayUs (50);

    zeiger++;
   }

  // Verbindung beenden
  i2c_close ();

  // Puffer umkopieren
  memcpy (puffer, puf1, laenge);

  return (TRUE);

 }


/**************************************************************************
 *
 * schreibt n (max. 63) Bytes in das I2C-EEPROM an der bergebenen Adresse
 * Liefert FALSE, falls ein Fehler aufgetreten ist
 *
 * Wird die Seitengrenze von 64 Bytes berschritten, werden zwei
 * Schreibvorgnge notwendig
 *
 **************************************************************************/
char i2c_EEPROM_Write (char dev, unsigned int Adresse, char laenge, char* puffer)
 {

  // #define Pagesize  64              // Seitengre des EEPROMs

  char b;                           // Hi-Lowbyte-Split der Adresse
  char puf2 [Buflen];               // bergabepuffer
  char zeiger;                      // Zeiger auf bergabepuffer
  char tester;                      // Flag fr erfolgreiche bertragung
  unsigned int Adresszaehler;       // Mitgefhrter Adresszhler

  // Puffer umkopieren
  Adresszaehler = Adresse;
  memset (puf2, 255, Buflen);
  memcpy (puf2, puffer, laenge);

// print (UART1, "A");

  // Slave adressieren zum Schreiben
  if (!(i2c_open (dev, WRITE)))
   {
     // Verbindung beenden
     i2c_close ();
     // print (UART1, "R");

     return (FALSE);
   }

// print (UART1, "B");

  // Adresse schreiben
  // High-Byte
  b = (Adresse & 0xff00) >> 8;                            // Highbyte ausmaskieren
  if (!(i2c_writebyte (b)))
   {
     // Verbindung beenden
     i2c_close ();
     return (FALSE);
   }

  // Low-Byte
  b = (Adresse & 0xff);                                   // Lowbyte ausmaskieren
  if (!(i2c_writebyte (b)))
   {
     // Verbindung beenden
     i2c_close ();
     return (FALSE);
   }

// print (UART1, "C");

  // Daten schreiben
  zeiger = 0;
  while (zeiger < laenge)
   {
    tester = i2c_writebyte (puf2 [zeiger]);           // Datum schreiben

    if (!(tester))
     {
      // Verbindung beenden
      i2c_close ();
      return (FALSE);
     }

    DelayUs (100);                                   // Ein bischen Zeit verbummeln

    //--------------------------------------------
    // berprfung auf berschrittene Seitengrenze
    //--------------------------------------------
    if ( ((Adresszaehler+1) % 64) == 0)
     { // Seitengrenze berschritten

       //--------------------------------------------------------------------------
       // Hier wird es notwendig, den Schreibvorgang abzuschliessen und das EEPROM
       // anschliessend neu zu adressieren (warum steht das nicht im Datenblatt ???)
       //--------------------------------------------------------------------------

       // Verbindung beenden
       i2c_close ();

       Delayms (10);

       // berprfen, ob der Schreibzyklus abgelaufen ist. Erst dann gibt es wieder ein ACK
       tester = FALSE;
       while (!(tester))
        {
         tester = i2c_open (dev, WRITE);
         DelayUs (100);
         i2c_close ();
         DelayUs (100);
        }

       // Neu Adressieren zum Schreiben
       i2c_open (dev, WRITE);

       // Adresse neu setzen
       b = ((Adresszaehler+1) & 0xff00) >> 8;                            // Highbyte ausmaskieren
       i2c_writebyte (b);
       b = ((Adresszaehler+1) & 0xff);                                   // Lowbyte ausmaskieren
       i2c_writebyte (b);

       //---------------------------------------------------
       // Neuadressierung ist abgeschlossen, weiter im Text
       //---------------------------------------------------

     } // Von if ( ((Adresszaehler+1) % Pagesize) == 0)

    zeiger++;                                        // Zeiger erhhen
    Adresszaehler++;                                 // Adresszhler mitfhren
   }

  //-------------------------
  // Daten schreiben beendet
  //-------------------------

// print (UART1, "D");

  // Verbindung beenden
  i2c_close ();

  Delayms (10);

  // berprfen, ob der Schreibzyklus abgelaufen ist. Erst dann gibt es wieder ein ACK
  tester = FALSE;
  while (!(tester))
   {
    tester = i2c_open (dev, WRITE);
    DelayUs (100);
    i2c_close ();
    DelayUs (100);
   }

  return (TRUE);
 }

/**************************************************************************
 *
 * Liefert True, wenn der Slave vorhanden ist
 *
 **************************************************************************/
char i2c_BusScan (char Adresse)
 {
  char Flag;

  Flag = i2c_open (Adresse, WRITE);         // Slave ansprechen
  DelayUs (1000);                           // Kurze Pause
  i2c_close ();                             // Verbindung wieder beenden
  return (Flag);                            // Ergebniss zurck
 }


/**************************************************************************************************************
 *
 * Ein Byte zum LCD schreiben im 4-Bit-Modus.
 *
 **************************************************************************************************************/
void lcd_write (unsigned char c)
 {
  LCD_RW = 0;
	PORTD = (PORTD & 0x0F) | (c & 0xF0);
	DelayUs (Pause);

	LCD_EN0 = 1;
	LCD_EN0 = 1;
	LCD_EN0 = 0;

	DelayUs (Pause);
	PORTD = (PORTD & 0x0F) |  (c << 4);    // Low-Nibble
	DelayUs (Pause);

	LCD_EN0 = 1;
	LCD_EN0 = 1;
  LCD_EN0 = 0;

	DelayUs (Pause);
 }

/**************************************************************************************************************
 *
 * Cursor ein/ausschalten
 *
 **************************************************************************************************************/
void lcd_cursor_on (char type)
 {

 	LCD_RS = 0;		                         // write control bytes
  if (type) lcd_write (0x0F);	else       // display on, blink curson on
  lcd_write (0x0C);	                   // display on, blink curson off

	Delayms (20);
 }

/**************************************************************************************************************
 *
 * LCD lschen und Cursor auf Null setzen
 *
 **************************************************************************************************************/
void lcd_clear (void)
 {
	LCD_RS = 0;
	lcd_write (0x1);
	Delayms (10);
 }


/**************************************************************************************************************
 *
 * String auf LCD ausgeben
 *
 **************************************************************************************************************/
void lcd_puts (const char * s)
 {
	LCD_RS = 1;	                                 // write characters
	while(*s)
	 {
	  lcd_write (*s);
	  // DelayUs (Pause);
	  s++;
	 }
 }

/***************************
 *
 * Positioniert den Cursor
 *
 ***************************/
void lcd_goto (unsigned char pos)
 {
	LCD_RS = 0;
	lcd_write (0x80+pos);
 }

/****************************************
 *
 * Text an definierter Position ausgeben
 *
 ****************************************/
void lcd_print (char x, char y, const char * s)
 {
  lcd_goto (x + (y*64));
  lcd_puts (s);
 }

/**************************************************************************************************************
 *
 * LCD initialisieren
 *
 **************************************************************************************************************/
void lcd_init (void)
 {
  GIE = 0;
  LCD_RW = 0;

	// Controller 1
	LCD_RS = 0;		    // write control bytes
	Delayms (20);		  // power on delay
	PORTD = 0x30;		  // attention!
	LCD_EN0 = 1;
	LCD_EN0 = 1;
  NOP();NOP();NOP();NOP();NOP();NOP();NOP();NOP();
	LCD_EN0 = 0;

	Delayms (10);
  LCD_EN0 = 1;
	LCD_EN0 = 1;
NOP();NOP();NOP();NOP();NOP();NOP();NOP();NOP();
	LCD_EN0 = 0;

	DelayUs (100);
	LCD_EN0 = 1;
	LCD_EN0 = 1;
NOP();NOP();NOP();NOP();NOP();NOP();NOP();NOP();
	LCD_EN0 = 0;

	Delayms (10);
	PORTD = 0x20;		  // set 4 bit mode
	LCD_EN0 = 1;
	LCD_EN0 = 1;
NOP();NOP();NOP();NOP();NOP();NOP();NOP();NOP();
	LCD_EN0 = 0;

	DelayUs (80);
	lcd_write (0x28);	// 4 bit mode, 1/16 duty, 5x8 font
	lcd_write (0x08);	// display off
	lcd_write (0x0C);	// display on, blink curson off
	lcd_write (0x06);	// entry mode
	Delayms (20);

  GIE = 1;
 }



/*****************************************************************************************************************
 *
 * UART initialisieren
 *
 *****************************************************************************************************************/
void UART_Init (void)
 {
  // UART-Einstellungen

  // Baudraten
  SPBRG1=16;       							                  // USART1: 19200 Baud bei 20MHZ - 8Bit-Mode mit BRGH=0
  SPBRG2=16;       							                  // USART2: 19200 Baud bei 20MHZ - 8Bit-Mode mit BRGH=0

  // Statusregister
  TXSTA1=0x24;									                    // UART1 : BRGH=1 / TXEN=1
  TXSTA2=0x24;									                    // UART2 : BRGH=1 / TXEN=1

  RCSTA1=0x90;                                      // UART1 : SPEN=1 / CREN=1
  RCSTA2=0x90;                                      // UART2 : SPEN=1 / CREN=1

  BRGH1 = 0;
  BRGH2 = 0;

  // Interrupts freigeben
  RC1IE=0;										                      // UART1 : Receive-Interrupt ein
  RC2IE=0;										                      // UART2 : Receive-Interrupt ein

  RC1IF=0;										                      // Flag=0;
  RC2IF=0;										                      // Flag=0;
 }

/***************************************************************************************************************
 *
 * ExaktePause
 *
 ***************************************************************************************************************/
void DelayMillisec (int x)
 {
  unsigned long tick;

  tick = Systemticker + (x / 10);
  while (tick > Systemticker);
 }


/***************
 *
 * Zentrale ISR
 *
 ***************/
interrupt void INT_(void)
 {

   //--------------------
   // Timer 1 berlauf ?
   //--------------------
   if (TMR1IF)
   	{

      //-------------------
      // Ablauf alle 10 ms
      //-------------------
      TMR1IF=0;                                         // Clear Interrupt-Flag
      TMR1H=0xE7;                                       // n = 2^16 - t [s] * 625000
      TMR1L=0x96;                                       // Timer 1 neu vorladen (0xE796 = 5 ms)

      //---------------
      // Zeitsteuerung
      //---------------
      Systemticker++;

      //-----------------
      // Sekunden zhlen
      //-----------------
      Sekundenticker++;
      if (Sekundenticker == 100)
       {
        Sekundenticker = 0;
        Sekundenzaehler++;
       }

   	} // From Timer 1 berlauf

   return;									                     // Zurck zum Hauptprogramm
 }

/***************************
 *
 * Initialisierungsroutinen
 *
 ***************************/
void sysinit (void)
 {

  //--------------------
  // IO-Ports klarmachen
  //--------------------

  // Alle Ports initialisieren bzw. auf Null setzen
  PORTA = 0;
  PORTB = 0;
  PORTC = 0;
  PORTD = 0;
  PORTE = 0;
  PORTF = 0;

  //-------------------------
  // Comperatoren abschalten
  //-------------------------
  CMCON = 0x07;

  // Ein- Ausgangsdefinitionen, 0 = Output / 1 Input, siehe Schaltplan
  TRISA = 0b11111111;  //
  TRISB = 0b11111111;  //
  TRISC = 0b10011000;  //
  TRISD = 0b00000000;  //
  TRISE = 0b11111110;  //
  TRISF = 0b11111111;  //
  TRISG = 0b11100100;  //

  //-----------------
  // UARTS klarmachen
  //-----------------
  UART_Init ();

  //-----------------------------
  // Glotzn uff du Arschkrampe !
  //-----------------------------
  lcd_init ();
  lcd_clear ();

  //----------------------
  // I2C startklar machen
  //----------------------
  i2c_init ();

  //-----------------------------
  // AD-Wandler startklar machen
  //-----------------------------
  adc_init ();

  //-------------------------------------
  // Timer 1 als Systemticker einrichten
  //-------------------------------------
  TMR1IF=0;                                         // Clear Interrupt-Flag
  TMR1H=0xE7;                                       // n = 2^16 - t [s] * 625000
  TMR1L=0x96;                                       // Timer 1 neu vorladen (0xE796 = 5 ms)
  TMR1IE=1;									                        // enable Interrupt of TMR1
  T1CON=49;									                        // Prescaler=1:8 + Timer on (49)

  //--------------------
  // Alle Interrupts an
  //--------------------
  PEIE=1;          							                    // Peripherie Int=on
  GIE=1;										                        // Define Interupts

  //----------------------
  // Begrungsbildschirm
  //----------------------
  Delayms (200);
  strcpy (outpbuf,"BumsWaage 1.0\0");
  lcd_print (1,0, outpbuf);
  Delayms (500);
  lcd_clear ();

  print (UART1, "\n\rBumswaage 1.0\n\r");
  print (UART1, "Start\n\r");

  //-------------------------
  // berprfen der Speicher
  //-------------------------
  print (UART1, "\n\rEEPROM 1 : ");
  if (i2c_BusScan (0xA0)) print (UART1, "OK\n\r"); else print (UART1, "Nicht OK\n\r");
  /*
  print (UART1, "EEPROM 2 : ");
  if (i2c_BusScan (0xA2)) print (UART1, "OK\n\r"); else print (UART1, "Nicht OK\n\r");
  */
 }


/**************************
 *
 * Tastenstatus reinholen
 *
 **************************/
char GetTaste (void)
 {
  return (PORTE);
 }

/**************************
 *
 * Tastenstatus reinholen
 *
 **************************/
char GetTasteLinks (void)
 {
  return (PORTE & 8);
 }

/**************************
 *
 * Tastenstatus reinholen
 *
 **************************/
char GetTasteRechts (void)
 {
  return (PORTE & 32);
 }

/**************************
 *
 * Tastenstatus reinholen
 *
 **************************/
char GetTasteMitte (void)
 {
  return (PORTE & 128);
 }


/************************************
 *
 * Liefert einen lesbaren Zahlenwert
 *
 ************************************/
void ZeigeErgebniss (int wert)
 {
  int erg;
  char puffer [10];

  memset (outpbuf, 0, Buflen);

  erg = wert;
  if (erg < 0)
   {
    outpbuf [0] = '-';
    erg *= (-1);
   } else outpbuf [0] = ' ';

 sprintf (puffer, "%2d", erg / 10);
 strcat (outpbuf, puffer);
 strcat (outpbuf, ".");
 sprintf (puffer, "%d", erg % 10);
 strcat (outpbuf, puffer);

 }

/*****************************************
 *
 * Einfache Zahleneingabe
 *
 *****************************************/
int Zahleneingabe (int wert)
 {
  int Rueckgabe;
  char Eingabe = TRUE;

  Rueckgabe = wert;

  ZeigeErgebniss (Rueckgabe);
  lcd_print (9,0, outpbuf);

  while (Eingabe)
   {
    //-------------------
    // Tastatur abfragen
    //-------------------
    if (GetTasteLinks ())
     {
      Delayms (10);
      while (GetTasteLinks  ());
      Delayms (10);
      Rueckgabe--;
      if (Rueckgabe < 0) Rueckgabe = 250;
      ZeigeErgebniss (Rueckgabe);
      lcd_print (9,0, outpbuf);
     }

    if (GetTasteMitte ())
     {
      Delayms (10);
      while (GetTasteMitte  ());
      Delayms (10);
      Rueckgabe++;
      if (Rueckgabe > 250) Rueckgabe = 0;
      ZeigeErgebniss (Rueckgabe);
      lcd_print (9,0, outpbuf);
     }

   if (GetTasteRechts ())
     {
      Delayms (10);
      while (GetTasteRechts  ());
      Delayms (10);
      Eingabe = FALSE;
     }


   } // von while (Eingabe)


  return (Rueckgabe);
 }

/*******************************************
 *
 * Macht eine kurze Pause und spricht dabei
 *
 *******************************************/
char WartePause (int wert)
 {
  char Pausenwert;
  char Ende = FALSE;

  // Startbedingungen klarmachen
  Pausenwert = wert / 100;
  SekundentickerAlt = 0;
  Sekundenzaehler = 0;

  // Text anzeigen
  sprintf (outpbuf,"Warte %d s     ", Pausenwert);
  lcd_print (0,1, outpbuf);

  // Zeit laufen lassen
  while ( (Pausenwert > 0) && (!Ende) )
   {

    // Uhr anzeigen
    if (SekundentickerAlt != Sekundenzaehler)
     {
      sprintf (outpbuf,"Warte %d s     ", Pausenwert);
      lcd_print (0,1, outpbuf);
      SekundentickerAlt = Sekundenzaehler;
      Pausenwert--;
     } // Von if (SekundentickerAlt != Sekundenzaehler)

    // Bei erneutem Tastendruck Funktion abbrechen
    if (GetTasteRechts ())
     {
      Ende = TRUE;
      Delayms (10);
      while (GetTasteRechts  ());
      Delayms (10);
     }

   } // Von while (Pausenwert > 0)

  return (Ende);

 }

/*******************
 *
 * Mach mal PIIIEP
 *
 *******************/
void Piep (int wert)
 {
   RC5=1;
   Delayms (wert);
   RC5=0;
 }

/***************************
 *
 * Zeichnet einen Kanal auf
 *
 ***************************/
void Aufzeichnungung (Slotnum)
 {
  char Abbruch = FALSE;
  char Start = FALSE;

  lcd_print (0,1, HilfeText2);
  while (!Abbruch)
   {

     while (!Start)
      {
       //----------------------
       // Tastenabfrage ESC
       //----------------------
       if (GetTasteLinks  ())
        {
         Delayms (10);
         while (GetTasteLinks  ());
         Delayms (10);
         Start = TRUE;
         Abbruch = TRUE;
        }

       //----------------------
       // Tastenabfrage Start
       //----------------------
       if (GetTasteRechts  ())
        {
         Delayms (10);
         while (GetTasteRechts  ());
         Delayms (10);
         Start = TRUE;
        }

      } // von while (!Start)

    // Messung beginnt
    if (!Abbruch)
     {
      Start = FALSE;

      //----------------------
      // Ablauf der Wartezeit
      //----------------------
      if (WartePause (PreTimer)) Abbruch = TRUE;

      //--------------------
      // Ablauf der Messung
      //--------------------
      if (!Abbruch)
       {
        Piep (20);
        Piep (20);
        Piep (20);
        Piep (20);
        Piep (20);
        Piep (20);

        // Vorbereitungen zur Messung
        Ticker = Systemticker;
        Sekundenzaehler = 0;
        Sekundenticker = 0;
        SekundentickerAlt = 0;
        MessPufferZeiger = 0;

        sprintf (outpbuf,"Speichere    ");
        lcd_print (0,1, outpbuf);

        while ( ((Ticker+Aufzeichnung) > Systemticker) && (!Abbruch) )
         {

          // Synchro mit Ticker
          Pauser = Systemticker;

          // Daten reinholen und im internen RAM ablegen.
          Messpuffer [MessPufferZeiger] = adc_get_newton ();

          MessPufferZeiger++;

          // Uhr anzeigen
          if (SekundentickerAlt != Sekundenzaehler)
           {
            sprintf (outpbuf,"%d s", Sekundenzaehler);
            lcd_print (11,1, outpbuf);
            SekundentickerAlt = Sekundenzaehler;
           } // Von if (SekundentickerAlt != Sekundenzaehler)

          // Bei erneutem Tastendruck Funktion abbrechen
          if (GetTasteRechts ())
           {
            Abbruch = TRUE;
            Delayms (10);
            while (GetTasteRechts  ());
            Delayms (10);
           }

          // Warten bis Zeit wieder abgelaufen ist
          while (Pauser == Systemticker);

         } // Von while ((Ticker+Aufzeichnung) < Systemticker)

        sprintf (outpbuf,"                ");
        lcd_print (0,1, outpbuf);

       }

      //----------------------------
      // Speicherung der Ergebnisse
      //----------------------------
      if (!Abbruch)
       {

        DevAdresse = 0xA0;
        Adresse = Slotnum * Aufzeichnung * 2;
        RAMZeiger = 0;

        sprintf (outpbuf, "Slot %d EEPROM %d Adresse %d\n\r", Slotnum, DevAdresse, Adresse);
        print (UART1, outpbuf);

        sprintf (outpbuf,"Sicherung       ");
        lcd_print (0,1, outpbuf);
        MessPufferZeiger = 0;
        Zeiger = &Messpuffer [0];
        for (i=0; i < Aufzeichnung; i += 5)
         {
          i2c_EEPROM_Write (DevAdresse, Adresse, 10, Zeiger); // Immer 10 Bytes speichern

          Adresse += 10;
          Zeiger += 10;

          // Anzeige des Fortschritts
          if ((Adresse % 20) == 0)
           {
            sprintf (outpbuf, "%d ", Adresse);
            lcd_print (10,1, outpbuf);
           }

          // Debugausgabe
          sprintf (outpbuf, "%d ", Messpuffer [i]);
          print (UART1, outpbuf);

         }  // Von for (i=0; i < Aufzeichnung; i += 5)

        sprintf (outpbuf,"                ");
        lcd_print (0,1, outpbuf);

        Piep (100);

       }

      //---------------
      // Fertigmeldung
      //---------------
      if (!Abbruch)
       {
        lcd_print (0,1, HilfeText3);
        while (!(GetTasteRechts ()));
        Delayms (10);
        while (GetTasteRechts  ());
        Delayms (10);
        sprintf (outpbuf,"                ");
        lcd_print (0,1, outpbuf);
       }

     } // Von if (!Abbruch)

    // Textzeile wiederherstellen
    lcd_print (0,1, HilfeText2);

   }  // Von while (!Abbruch)

 }

/****************************
 *
 * Analysiert Aufzeichnungen
 *
 ****************************/
void Analyse (Slotnum)
 {
   int min;
   int max;
   int avg;
   long lavg;
   char Ende = FALSE;
   char cdummy;
   int wert;
   float fImpuls = 0;
   int p;

   lcd_print (0,1, HilfeText5);

   //----------------------------
   // Daten aus dem EEPROM laden
   //----------------------------
   sprintf (outpbuf, "\n\rAnalyse Slot %d\n\r", Slotnum+1);
   print (UART1, outpbuf);
   DevAdresse = 0xA0;
   Adresse = Slotnum * Aufzeichnung * 2;
   MessPufferZeiger = 0;

   sprintf (outpbuf, "Slot %d EEPROM %d Adresse %d\n\r", Slotnum, DevAdresse, Adresse);
   print (UART1, outpbuf);

   for (i=0; i < Aufzeichnung; i++)
    {
     i2c_EEPROM_Read (0xA0, Adresse, 2, outpbuf);
     Messpuffer [i] = (outpbuf [0] + outpbuf [1]*256);
     Adresse += 2;
    }

   //-----------------------
   // Min/Max/Avg berechnen
   //-----------------------
   min = Messpuffer [0];
   max = Messpuffer [0];
   lavg = Messpuffer [0];

   for (i=1; i < Aufzeichnung; i++)
    {
     if (Messpuffer [i] < min) min = Messpuffer [i];
     if (Messpuffer [i] > max) max = Messpuffer [i];
     lavg += Messpuffer [i];
     fImpuls += 0.01*Messpuffer [i];
    }
   avg = (int)(lavg / Aufzeichnung);
   p = (int)(fImpuls);

   // Debugausgabe
   print (UART1, "Min : ");
   ZeigeErgebniss (min);
   print (UART1, outpbuf);
   print (UART1, "N  Max : ");
   ZeigeErgebniss (max);
   print (UART1, outpbuf);
   print (UART1, "N  Avg : ");
   ZeigeErgebniss (avg);
   print (UART1, outpbuf);
   print (UART1, "N  p : ");
   ZeigeErgebniss (p);
   print (UART1, outpbuf);
   print (UART1, "Ns\n\r");

   // Maximalwert ausgeben
   lcd_print (0,1, HilfeText4);
   ZeigeErgebniss (max);
   strcat (outpbuf, " N");
   lcd_print (3,1, outpbuf);

   //------------------------------------------
   // Warten auf Besttigung oder das Auslesen
   //------------------------------------------
   while (RC1IF)
    {
     Delayms (20);
     RC1IF = 0;
    }
   while (!Ende)
    {
      // Taste abfragen
      if (GetTasteRechts  ())
       {
         Delayms (10);
         while (GetTasteRechts  ());
         Delayms (10);
         Ende = TRUE;
       }

      // Daten bertragen ?
      if (RC1IF)
       {
        cdummy = RCREG1;
        sprintf (outpbuf, "\n\rSlot %d\n\r", Slotnum+1);
        print (UART1, outpbuf);
        DevAdresse = 0xA0;
        Adresse = Slotnum * Aufzeichnung * 2;

        sprintf (outpbuf, "Slot %d EEPROM %d Adresse %d\n\r", Slotnummer, DevAdresse, Adresse);
        print (UART1, outpbuf);

        for (i=0; i < Aufzeichnung; i++)
         {

          i2c_EEPROM_Read (0xA0, Adresse, 2, outpbuf);
          wert = (outpbuf [0] + outpbuf [1]*256);
          Adresse += 2;

          sprintf (outpbuf, "%4d ", i*10);
          print (UART1, outpbuf);
          ZeigeErgebniss (wert);
          print (UART1, outpbuf);
          print (UART1, "\r");

         }

        // Auslesen beendet
        RC1IF = 0;

       }

    } // Von  while (!Ende)

   sprintf (outpbuf,"                ");
   lcd_print (0,1, outpbuf);

 }

/***************************
 *
 * Momentanen Wert anzeigen
 *
 ***************************/
void ZeigeAktuell (void)
 {
  char Ende = FALSE;
  int wert = 0;

  lcd_print (0,1, HilfeText6);
  while (!Ende)
   {
      // Taste abfragen
      if (GetTasteRechts  ())
       {
         Delayms (10);
         while (GetTasteRechts  ());
         Delayms (10);
         Ende = TRUE;
       }

     // Wert anzeigen
     wert = adc_get_raw ();
     sprintf (outpbuf, "%4d=", wert);
     lcd_print (0,1, outpbuf);
     wert = adc_get_newton ();
     ZeigeErgebniss (wert);
     lcd_print (5,1, outpbuf);
     lcd_print (10,1, "N");
     Delayms (200);


   } // Von while (!Ende)

 }

/**********************
 *
 * Messung Kalibrieren
 *
 **********************/
void Kalibration (void)
 {
  char Ende = FALSE;

  //-------------------
  // Sicherheitsabfrage
  //-------------------
  lcd_print (0,1, HilfeText9);
  while (!Ende)
   {
      // Taste abfragen
      if (GetTasteRechts  ())
       {
         Delayms (10);
         while (GetTasteRechts  ());
         Delayms (10);
         Ende = TRUE;
       }

      // Taste abfragen
      if (GetTasteLinks  ())
       {
         Delayms (10);
         while (GetTasteLinks  ());
         Delayms (10);
         return;
       }

      // Taste abfragen
      if (GetTasteMitte  ())
       {
         Delayms (10);
         while (GetTasteMitte  ());
         Delayms (10);
         return;
       }

   } // Von while (!Ende)

  //---------------------
  // Kalibration startet
  //---------------------

  // Punkte aus dem EEP laden
  x1 = EEPRead (0) * 256;
  x1 += EEPRead (1);
  y1 = EEPRead (2) * 256;
  y1 += EEPRead (3);
  x2 = EEPRead (4) * 256;
  x2 += EEPRead (5);
  y2 = EEPRead (6) * 256;
  y2 += EEPRead (7);

  // Korrekturwerte setzen
  if (x1 > 1023) x1 = 10;
  if (y1 > 150) y1 = 100;
  if (x2 > 1023) x2 = 1000;
  if (y2 > 150) y2 = 100;

  //-------------------
  // Punkt 1 bestimmen
  //-------------------
  lcd_print (0,0, HilfeText7); // "Punkt 1 : __._N"
  lcd_print (0,1, HilfeText8); // "<->   <+>  <OK>"
  y1 = Zahleneingabe (y1);
  x1 = adc_get_raw ();

  //-------------------
  // Punkt 2 bestimmen
  //-------------------
  lcd_print (6,0, "2");
  lcd_print (0,1, HilfeText8); // "<->   <+>  <OK>"
  y2 = Zahleneingabe (y2);
  x2 = adc_get_raw ();

  //----------------------
  // Ergebnisse speichern
  //----------------------
  EEPWrite (0, (char)(x1 >> 8));
  EEPWrite (1, (char)(x1 & 0x00FF));
  EEPWrite (2, (char)(y1 >> 8));
  EEPWrite (3, (char)(y1 & 0x00FF));
  EEPWrite (4, (char)(x2 >> 8));
  EEPWrite (5, (char)(x2 & 0x00FF));
  EEPWrite (6, (char)(y2 >> 8));
  EEPWrite (7, (char)(y2 & 0x00FF));

 }

/**************************
 *
 * Baut das Hauptmen auf
 *
 **************************/
void HauptMenue ()
 {
   char Hauptpos = 0;

   // Ersten Eintrag darstellen
   lcd_print (0,0, HauptMenueText [0]);
   lcd_print (0,1, HilfeText1);

   while (1)
    {

     //----------------------
     // Tastenabfrage Weiter
     //----------------------
     if (GetTasteLinks  ())
      {
       Delayms (10);
       while (GetTasteLinks  ());
       Delayms (10);
       Hauptpos++;
       if (Hauptpos == HauptMenueAnzahl) Hauptpos = 0;
       lcd_print (0,0, HauptMenueText [Hauptpos]);
       lcd_print (0,1, HilfeText1);
      }

     //-------------------------
     // Tastenabfrage Ja
     //-------------------------
     if (GetTasteRechts  ())
      {
       Delayms (10);
       while (GetTasteRechts  ());
       Delayms (10);

       // Auswahl der Sub-Funktionen
       if (Hauptpos == 0) Aufzeichnungung (0);		// Slot 1 Aufzeichnung
       if (Hauptpos == 1) Aufzeichnungung (1);		// Slot 2 Aufzeichnung
       if (Hauptpos == 2) Aufzeichnungung (2);		// Slot 3 Aufzeichnung
       if (Hauptpos == 3) Aufzeichnungung (3);		// Slot 4 Aufzeichnung

       if (Hauptpos == 4) Analyse (0);				// Slot 1 Analyse
       if (Hauptpos == 5) Analyse (1);				// Slot 2 Analyse
       if (Hauptpos == 6) Analyse (2);				// Slot 3 Analyse
       if (Hauptpos == 7) Analyse (3);				// Slot 4 Analyse

       if (Hauptpos == 8) ZeigeAktuell ();		// Momentanen Wert anzeigen

       if (Hauptpos == 9) Kalibration ();		  // Messung Kalibrieren

       // Men wieder aufbauen
       lcd_print (0,0, HauptMenueText [Hauptpos]);
       lcd_print (0,1, HilfeText1);

      }

    } // Von while (1)

 }

/*************
 *
 * Main
 *
 *************/
void main (void)
{
  //----------------------
  // System initialisieren
  //----------------------
  sysinit ();

  //-------------------------
  // Glaubst du an den User ?
  //-------------------------
  HauptMenue ();

}

// Thats all folks !


